@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
 # The MIT License (MIT)
-# Copyright (c) 2014 Matias Bordese
+# Copyright (c) 2014-2017 Matias Bordese
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -38,6 +38,7 @@ from unidiff.constants import (
     LINE_TYPE_NO_NEWLINE,
     LINE_VALUE_NO_NEWLINE,
     RE_HUNK_BODY_LINE,
+    RE_HUNK_EMPTY_BODY_LINE,
     RE_HUNK_HEADER,
     RE_SOURCE_FILENAME,
     RE_TARGET_FILENAME,
@@ -62,6 +63,7 @@ else:
     make_str = str
     implements_to_string = lambda x: x
     unicode = str
+    basestring = str
 
 
 @implements_to_string
@@ -103,6 +105,23 @@ class Line(object):
         return self.line_type == LINE_TYPE_CONTEXT
 
 
+@implements_to_string
+class PatchInfo(list):
+    """Lines with extended patch info.
+
+    Format of this info is not documented and it very much depends on
+    patch producer.
+
+    """
+
+    def __repr__(self):
+        value = "<PatchInfo: %s>" % self[0].strip()
+        return make_str(value)
+
+    def __str__(self):
+        return ''.join(unicode(line) for line in self)
+
+
 @implements_to_string
 class Hunk(list):
     """Each of the modified blocks of a file."""
@@ -132,9 +151,11 @@ class Hunk(list):
         return make_str(value)
 
     def __str__(self):
-        head = "@@ -%d,%d +%d,%d @@ %s\n" % (
+        # section header is optional and thus we output it only if it's present
+        head = "@@ -%d,%d +%d,%d @@%s\n" % (
             self.source_start, self.source_length,
-            self.target_start, self.target_length, self.section_header)
+            self.target_start, self.target_length,
+            ' ' + self.section_header if self.section_header else '')
         content = ''.join(unicode(line) for line in self)
         return head + content
 
@@ -169,9 +190,10 @@ class Hunk(list):
 class PatchedFile(list):
     """Patch updated file, it is a list of Hunks."""
 
-    def __init__(self, source='', target='',
+    def __init__(self, patch_info=None, source='', target='',
                  source_timestamp=None, target_timestamp=None):
         super(PatchedFile, self).__init__()
+        self.patch_info = patch_info
         self.source_file = source
         self.source_timestamp = source_timestamp
         self.target_file = target
@@ -181,10 +203,16 @@ class PatchedFile(list):
         return make_str("<PatchedFile: %s>") % make_str(self.path)
 
     def __str__(self):
-        source = "--- %s\n" % self.source_file
-        target = "+++ %s\n" % self.target_file
+        # patch info is optional
+        info = '' if self.patch_info is None else str(self.patch_info)
+        source = "--- %s%s\n" % (
+            self.source_file,
+            '\t' + self.source_timestamp if self.source_timestamp else '')
+        target = "+++ %s%s\n" % (
+            self.target_file,
+            '\t' + self.target_timestamp if self.target_timestamp else '')
         hunks = ''.join(unicode(hunk) for hunk in self)
-        return source + target + hunks
+        return info + source + target + hunks
 
     def _parse_hunk(self, header, diff, encoding):
         """Parse hunk details."""
@@ -200,7 +228,11 @@ class PatchedFile(list):
         for diff_line_no, line in diff:
             if encoding is not None:
                 line = line.decode(encoding)
-            valid_line = RE_HUNK_BODY_LINE.match(line)
+
+            valid_line = RE_HUNK_EMPTY_BODY_LINE.match(line)
+            if not valid_line:
+                valid_line = RE_HUNK_BODY_LINE.match(line)
+
             if not valid_line:
                 raise UnidiffParseError('Hunk diff line expected: %s' % line)
 
@@ -225,6 +257,11 @@ class PatchedFile(list):
             else:
                 original_line = None
 
+            # stop parsing if we got past expected number of lines
+            if (source_line_no > expected_source_end or
+                    target_line_no > expected_target_end):
+                raise UnidiffParseError('Hunk is longer than expected')
+
             if original_line:
                 original_line.diff_line_no = diff_line_no
                 hunk.append(original_line)
@@ -234,6 +271,11 @@ class PatchedFile(list):
                     target_line_no == expected_target_end):
                 break
 
+        # report an error if we haven't got expected number of lines
+        if (source_line_no < expected_source_end or
+                target_line_no < expected_target_end):
+            raise UnidiffParseError('Hunk is shorter than expected')
+
         self.append(hunk)
 
     def _add_no_newline_marker_to_last_hunk(self):
@@ -244,6 +286,12 @@ class PatchedFile(list):
         last_hunk.append(
             Line(LINE_VALUE_NO_NEWLINE + '\n', line_type=LINE_TYPE_NO_NEWLINE))
 
+    def _append_trailing_empty_line(self):
+        if not self:
+            raise UnidiffParseError('Unexpected trailing newline character')
+        last_hunk = self[-1]
+        last_hunk.append(Line('\n', line_type=LINE_TYPE_EMPTY))
+
     @property
     def path(self):
         """Return the file path abstracted from VCS."""
@@ -251,10 +299,10 @@ class PatchedFile(list):
                 self.target_file.startswith('b/')):
             filepath = self.source_file[2:]
         elif (self.source_file.startswith('a/') and
-                self.target_file == '/dev/null'):
+              self.target_file == '/dev/null'):
             filepath = self.source_file[2:]
         elif (self.target_file.startswith('b/') and
-                self.source_file == '/dev/null'):
+              self.source_file == '/dev/null'):
             filepath = self.target_file[2:]
         else:
             filepath = self.source_file
@@ -294,6 +342,11 @@ class PatchSet(list):
 
     def __init__(self, f, encoding=None):
         super(PatchSet, self).__init__()
+
+        # convert string inputs to StringIO objects
+        if isinstance(f, basestring):
+            f = self._convert_string(f, encoding)
+
         # make sure we pass an iterator object to parse
         data = iter(f)
         # if encoding is None, assume we are reading unicode data
@@ -303,15 +356,17 @@ class PatchSet(list):
         return make_str('<PatchSet: %s>') % super(PatchSet, self).__repr__()
 
     def __str__(self):
-        return '\n'.join(unicode(patched_file) for patched_file in self)
+        return ''.join(unicode(patched_file) for patched_file in self)
 
     def _parse(self, diff, encoding):
         current_file = None
+        patch_info = None
 
         diff = enumerate(diff, 1)
         for unused_diff_line_no, line in diff:
             if encoding is not None:
                 line = line.decode(encoding)
+
             # check for source file header
             is_source_filename = RE_SOURCE_FILENAME.match(line)
             if is_source_filename:
@@ -329,9 +384,11 @@ class PatchSet(list):
                 target_file = is_target_filename.group('filename')
                 target_timestamp = is_target_filename.group('timestamp')
                 # add current file to PatchSet
-                current_file = PatchedFile(source_file, target_file,
-                                           source_timestamp, target_timestamp)
+                current_file = PatchedFile(
+                    patch_info, source_file, target_file,
+                    source_timestamp, target_timestamp)
                 self.append(current_file)
+                patch_info = None
                 continue
 
             # check for hunk header
@@ -340,6 +397,7 @@ class PatchSet(list):
                 if current_file is None:
                     raise UnidiffParseError('Unexpected hunk found: %s' % line)
                 current_file._parse_hunk(line, diff, encoding)
+                continue
 
             # check for no newline marker
             is_no_newline = RE_NO_NEWLINE_MARKER.match(line)
@@ -347,6 +405,18 @@ class PatchSet(list):
                 if current_file is None:
                     raise UnidiffParseError('Unexpected marker: %s' % line)
                 current_file._add_no_newline_marker_to_last_hunk()
+                continue
+
+            # sometimes hunks can be followed by empty lines
+            if line == '\n' and current_file is not None:
+                current_file._append_trailing_empty_line()
+                continue
+
+            # if nothing has matched above then this line is a patch info
+            if patch_info is None:
+                current_file = None
+                patch_info = PatchInfo()
+            patch_info.append(line)
 
     @classmethod
     def from_filename(cls, filename, encoding=DEFAULT_ENCODING, errors=None):
@@ -355,13 +425,17 @@ class PatchSet(list):
             instance = cls(f)
         return instance
 
-    @classmethod
-    def from_string(cls, data, encoding=None, errors='strict'):
-        """Return a PatchSet instance given a diff string."""
+    @staticmethod
+    def _convert_string(data, encoding=None, errors='strict'):
         if encoding is not None:
             # if encoding is given, assume bytes and decode
             data = unicode(data, encoding=encoding, errors=errors)
-        return cls(StringIO(data))
+        return StringIO(data)
+
+    @classmethod
+    def from_string(cls, data, encoding=None, errors='strict'):
+        """Return a PatchSet instance given a diff string."""
+        return cls(cls._convert_string(data, encoding, errors))
 
     @property
     def added_files(self):
